from __future__ import absolute_import
import requests


class Remote(object):

    def __init__(self, hostname, protocol):
        """
        Arguments:
            `hostname`:     Hostname (e.g. google.com)
            `protocol`:     HTTP protocol (e.g. https)
        """
        self.protocol = protocol
        self.hostname = hostname

    def url(self, endpoint=''):
        """
        Get the base URL of the Remote.

        Arguments:
            None
        Returns:
            `str` base URL
        """
        if not endpoint.startswith('/'):
            endpoint = "/" + endpoint
        return self.protocol + "://" + self.hostname + endpoint

    def ping(self, endpoint=''):
        """
        Ping the server to make sure that you can access the base URL.

        Arguments:
            None
        Returns:
            `boolean` Successful access of server (or status code)
        """
        r = requests.get(self.url() + "/" + endpoint)
        return r.status_code



from __future__ import absolute_import
from .enums import *
from .errors import *
import numpy

from .RAMONBase import RAMONBase


class RAMONVolume(RAMONBase):
    """
    RAMONVolume Object for storing neuroscience data with a voxel volume
    """
    def __init__(self,
                 xyz_offset=(0, 0, 0),
                 resolution=0,
                 cutout=None,
                 voxels=None,

                 id=DEFAULT_ID,
                 confidence=DEFAULT_CONFIDENCE,
                 dynamic_metadata=DEFAULT_DYNAMIC_METADATA,
                 status=DEFAULT_STATUS,
                 author=DEFAULT_AUTHOR):
        """
        Initialize a new RAMONVolume object. Inherits attributes from RAMONBase
        as well as:

        Arguments:
            xyz_offset (int[3] : (0, 0, 0)): x,y,z coordinates of the minimum
                corner of the cube (if data is a cutout), otherwise empty
            resolution (int : 0): level in the database resolution hierarchy
            cutout (numpy.ndarray): dense matrix of data
            voxels: Unused for now
        """

        self.xyz_offset = xyz_offset
        self.resolution = resolution
        self.cutout = cutout
        self.voxels = voxels

        RAMONBase.__init__(self, id=id, confidence=confidence,
                           dynamic_metadata=dynamic_metadata,
                           status=status, author=author)

    def data(self):
        """
        Gets the data from the volume and pumps it into a numpy.ndarray format,
        regardless of whether it's stored in `cutout` or `voxels`. Returns it
        as though it were stored in `cutout`.

        This is useful for cases where you need to operate on a 3D matrix.

        Arguments:
            None

        Returns:
            numpy.ndarray
        """
        if self.cutout:
            return self.cutout
        else:
            raise NotImplementedError("Cannot convert from voxel list yet.")



from __future__ import absolute_import
import mcubes


def export_to_dae(filename, data, level=0):
    """
    Converts a dense annotation to a DAE, using Marching Cubes (PyMCubes).

    Arguments:
        filename (str): The filename to write out to
        data (numpy.ndarray): The dense annotation
        level (int): The level at which to run mcubes

    Returns:
        boolean success
    """

    if ".dae" not in filename:
        filename = filename + ".dae"

    vs, fs = mcubes.marching_cubes(cutout, level)
    mcubes.export_mesh(vs, fs, filename, "ndioexport")



from __future__ import absolute_import
import requests
import json

from .Remote import Remote
from .errors import *


DEFAULT_HOSTNAME = "lims.neurodata.io"
DEFAULT_PROTOCOL = "http"


class OCPMeta(Remote):
    """
    OCPMeta Remotes enable access to the Metadata-OCP API endpoints (which
    can be found at `api.neurodata.io/metadata/ocp/`). This class is useful
    for reading metadata for existing projects, or, if you specify a key,
    adding new metadata, manipulating your existing metadata, and archiving
    old metadata. Archived metadata is still stored by the LIMS system.
    """

    def __init__(self,
                 hostname=DEFAULT_HOSTNAME,
                 protocol=DEFAULT_PROTOCOL):
        super(OCPMeta, self).__init__(hostname, protocol)

    def get_metadata(self, token):
        """
        Get metadata via a project token.

        Arguments:
            token (str):      The project (token) to access

        Returns:
            JSON metadata associated with this project
        """

        req = requests.get(self.url("/metadata/ocp/get/" + token))
        return req.json()

    def set_metadata(self, token, data):
        """
        Insert new metadata into the OCP metadata database.

        Arguments:
            token (str): Token of the datum to set
            data (str): A dictionary to insert as metadata. Include `secret`.

        Returns:
            JSON of the inserted ID (convenience) or an error message.

        Throws:
            RemoteDataUploadError: If the token is already populated, or if
                there is an issue with your specified `secret` key.
        """
        req = requests.post(self.url("/metadata/ocp/set/" + token), json=data)

        if req.status_code != 200:
            raise RemoteDataUploadError(
                "Could not upload metadata: " + req.json()['message']
            )
        return req.json()




class InvalidRAMONError(Exception):
    """
    Raised when an invalid RAMON object is referenced or created.
    """
    def __init__(self, msg):
        self.msg = msg

    def __str__(self):
        return self.msg



from __future__ import absolute_import
from ndio.ramon.RAMONBase import *
from ndio.ramon.RAMONGeneric import *
from ndio.ramon.RAMONNeuron import *
from ndio.ramon.RAMONOrganelle import *
from ndio.ramon.RAMONSegment import *
from ndio.ramon.RAMONSynapse import *
from ndio.ramon.RAMONVolume import *

from .errors import *

"""

The enumerables in this file are to afford compatibility with CAJAL, which uses
integers as types. To convert from integer to string (for instance, if you want
to specify the ANNOTATION_TYPE in a RAMON hdf5 file), use:

    >>> AnnotationType.get_str(1)
    "GENERIC"

To convert from an integer to an actual RAMON type (say, if you want to create
a new RAMON object dynamically), use:

    >>> AnnotationType.get_class(1)
    ndio.ramon.RAMONGeneric.RAMONGeneric

So you can create a new RAMON object of dynamic type like this:

    >>> anno_type = 5
    >>> r = AnnotationType.get_class(anno_type)()
    >>> type(r)
    ndio.ramon.RAMONNeuron.RAMONNeuron

NOTE! If you already have an HDF5 file that contains a RAMON object, it is far
easier to use the prebuilt `hdf5_to_ramon()` function (below).

    >>> import h5py
    >>> f = h5py.File('myfile.hdf5', 'r')
    >>> r = hdf5_to_ramon(f)

"""

_types = {
    "GENERIC": 1,
    "SYNAPSE": 2,
    "SEED": 3,
    "SEGMENT": 4,
    "NEURON": 5,
    "ORGANELLE": 6,
    "ATTRIBUTEDREGION": 7,
    "VOLUME": 8
}

_reverse_types = {v: k for k, v in list(_types.items())}

_ramon_types = {
    _types["GENERIC"]: RAMONGeneric,
    _types["SYNAPSE"]: RAMONSynapse,
    _types["SEED"]:  None,
    _types["SEGMENT"]: RAMONSegment,
    _types["NEURON"]: RAMONNeuron,
    _types["ORGANELLE"]: RAMONOrganelle,
    # _types["ATTRIBUTEDREGION"]: None,
    _types["VOLUME"]: RAMONVolume
}

_reverse_ramon_types = {v: k for k, v in list(_ramon_types.items())}


class AnnotationType:
    GENERIC = _types["GENERIC"]
    SYNAPSE = _types["SYNAPSE"]
    SEED = _types["SEED"]
    SEGMENT = _types["SEGMENT"]
    NEURON = _types["NEURON"]
    ORGANELLE = _types["ORGANELLE"]
    ATTRIBUTEDREGION = _types["ATTRIBUTEDREGION"]
    VOLUME = _types["VOLUME"]

    @staticmethod
    def get_str(type):
        return _reverse_types[type]

    @staticmethod
    def get_class(type):
        return _ramon_types[type]

    @staticmethod
    def get_int(type):
        return _reverse_ramon_types[type]


def hdf5_to_ramon(hdf5, anno_id=None):
    """
    Converts an HDF5 file to a RAMON object. Returns an object that is a child-
    -class of RAMON (though it's determined at run-time what type is returned).

    Accessing multiple IDs from the same file is not supported, because it's
    not dramatically faster to access each item in the hdf5 file at the same
    time It's semantically and computationally easier to run this function
    several times on the same file.

    Arguments:
        hdf5 (h5py.File): A h5py File object that holds RAMON data
        anno_id (int): The ID of the RAMON obj to extract from the file. This
            defaults to the first one (sorted) if none is specified.

    Returns:
        ndio.RAMON object
    """

    if anno_id is None:
        # The user just wants the first item we find, so... Yeah.
        return hdf5_to_ramon(hdf5, list(hdf5.keys())[0])

    # First, get the actual object we're going to download.
    anno_id = str(anno_id)
    if anno_id not in list(hdf5.keys()):
        raise ValueError("ID {} is not in this file. Options are: {}".format(
            anno_id,
            ", ".join(list(hdf5.keys()))
        ))

    anno = hdf5[anno_id]
    # anno now holds just the RAMON of interest

    # This is the most complicated line in here: It creates an object whose
    # type is conditional on the ANNOTATION_TYPE of the hdf5 object.
    try:
        r = AnnotationType.get_class(anno['ANNOTATION_TYPE'][0])()
    except:
        raise InvalidRAMONError("This is not a valid RAMON type.")

    # All RAMON types definitely have these attributes:
    metadata = anno['METADATA']
    r.author = metadata['AUTHOR'][0]
    r.confidence = metadata['CONFIDENCE'][0]
    r.status = metadata['STATUS'][0]
    r.id = anno_id

    # These are a little tougher, some RAMON types have special attributes:

    if type(r) in [RAMONNeuron, RAMONSynapse]:
        r.segments = metadata['SEGMENTS'][()]

    if issubclass(type(r), RAMONVolume):
        r.cutout = anno['CUTOUT'][()]
        r.xyz_offset = anno['XYZOFFSET'][()]
        r.resolution = anno['RESOLUTION'][0]

    if type(r) is RAMONSynapse:
        r.synapse_type = metadata['SYNAPSETYPE'][0]
        r.weight = metadata['WEIGHT'][0]

    if type(r) is RAMONSegment:
        if 'NEURON' in metadata:
            r.neuron = metadata['NEURON'][0]
        if 'PARENTSEED' in metadata:
            r.parent_seed = metadata['PARENTSEED'][0]
        if 'SEGMENTCLASS' in metadata:
            r.segment_class = metadata['SEGMENTCLASS'][0]
        if 'SYNAPSES' in metadata:
            r.synapses = metadata['SYNAPSES'][()]
        if 'ORGANELLES' in metadata:
            r.organelles = metadata['ORGANELLES'][()]

    if type(r) is RAMONOrganelle:
        r.organelle_class = metadata['ORGANELLECLASS'][0]

    return r


def ramon_to_hdf5(ramon, hdf5=None):
    """
    Exports a RAMON object to an HDF5 file object.

    Arguments:
        ramon (RAMON): A subclass of RAMONBase
        hdf5 (str): Export filename

    Returns:
        hdf5.File

    Raises:
        InvalidRAMONError: if you pass a non-RAMON object
    """

    if issubclass(type(ramon), RAMONBase) is False:
        raise InvalidRAMONError("Invalid RAMON supplied to ramon_to_hdf5.")

    import h5py
    import numpy

    if hdf5 is None:
        with h5py.File('{}.hdf5'.format(ramon.id), 'a') as hdf5:

            # First we'll export things that all RAMON objects have in common,
            # starting with the Group that encompasses each ID:
            grp = hdf5.create_group(str(ramon.id))

            grp.create_dataset("ANNOTATION_TYPE", (1,),
                               numpy.uint32,
                               data=AnnotationType.get_int(type(ramon)))
            if hasattr(ramon, 'resolution'):
                grp.create_dataset('RESOLUTION', (1,),
                                   numpy.uint32, data=ramon.resolution)
            if hasattr(ramon, 'xyz_offset'):
                grp.create_dataset('XYZOFFSET', (3,),
                                   numpy.uint32, data=ramon.xyz_offset)
            if hasattr(ramon, 'cutout'):
                grp.create_dataset('CUTOUT', ramon.cutout.shape,
                                   ramon.cutout.dtype, data=ramon.cutout)

            # Next, add general metadata.
            metadata = grp.create_group('METADATA')

            metadata.create_dataset('AUTHOR', (1,),
                                    dtype=h5py.special_dtype(vlen=str),
                                    data=ramon.author)
            metadata.create_dataset('CONFIDENCE', (1,), numpy.float,
                                    data=ramon.confidence)
            metadata.create_dataset('STATUS', (1,), numpy.uint32,
                                    data=ramon.status)

            # Finally, add type-specific metadata:

            if hasattr(ramon, 'segments'):
                # metadata.create_dataset('SEGMENTS', (len(ramon.segments), 2),
                #                         numpy.uint32, data=ramon.segments)
                metadata.create_dataset('SEGMENTS', data=numpy.ndarray(ramon.segments))

            if hasattr(ramon, 'synapse_type'):
                metadata.create_dataset('SYNAPSETYPE', (1,), numpy.uint32,
                                        data=ramon.synapse_type)

            if hasattr(ramon, 'weight'):
                metadata.create_dataset('WEIGHT', (1,),
                                        numpy.float, data=ramon.weight)

            if hasattr(ramon, 'neuron'):
                metadata.create_dataset('NEURON', (1,),
                                        numpy.uint32, data=ramon.neuron)

            if hasattr(ramon, 'segment_class'):
                metadata.create_dataset('SEGMENTCLASS', (1,), numpy.uint32,
                                        data=ramon.segment_class)

            if hasattr(ramon, 'synapses'):
                metadata.create_dataset('SYNAPSES', (len(ramon.synapses),),
                                        numpy.uint32, data=ramon.synapses)

            if hasattr(ramon, 'organelles'):
                metadata.create_dataset('ORGANELLES', (len(ramon.organelles),),
                                        numpy.uint32, data=ramon.organelles)

            if hasattr(ramon, 'organelle_class'):
                metadata.create_dataset('ORGANELLECLASS', (1,), numpy.uint32,
                                        data=ramon.organelle_class)

            return hdf5



from __future__ import absolute_import
from .enums import *
from .errors import *
import numpy

from .RAMONBase import RAMONBase


class RAMONNeuron(RAMONBase):
    """
    RAMONNeuron Object for storing neuroscience data with a voxel volume
    """
    def __init__(self,
                 segments=None,

                 id=DEFAULT_ID,
                 confidence=DEFAULT_CONFIDENCE,
                 dynamic_metadata=DEFAULT_DYNAMIC_METADATA,
                 status=DEFAULT_STATUS,
                 author=DEFAULT_AUTHOR):

            self.segments = segments

            RAMONBase.__init__(self,
                                 id=id,
                                 confidence=confidence,
                                 dynamic_metadata=dynamic_metadata,
                                 status=status,
                                 author=author)



# Copyright 2014 Open Connectome Project (http://openconnecto.me)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import json
import argparse
import requests
import os
import requests


class AutoIngest:

    def __init__(self):
        self.channels
        self.dataset
        self.project
        self.metadata

    def add_channel(
        self, channel_name, datatype, channel_type, data_url, file_format,
            file_type, exceptions=0, resolution=0,
            windowrange=(0, 0), readonly=0):
        """
        Arguments:
            channel_name (str): Channel Name is the specific name of a
            specific series of data. Standard naming convention is to do
            ImageTypeIterationNumber or NameSubProjectName.

            datatype (str): The data type is the storage method of data in
            the channel. It can be uint8, uint16, uint32, uint64, or
            float32.

            channel_type (str): The channel type is the kind of data being
            stored in the channel. It can be image, annotation, or
            timeseries.

            data_url (str): This url points to the root directory of the
            files. Dropbox (or any data requiring authentication to
            download such as s3) is not an acceptable HTTP Server.

            file_format (str): File format refers to the overarching kind
            of data, as in slices (normal image data) or catmaid
            (tile-based).

            file_type (str): File type refers to the specific type of file
            that the data is stored in, as in, tiff, png, or tif.

            exceptions (int): Exceptions is an option to enable the
            possibility for annotations to contradict each other (assign
            different values to the same point). 1 corresponds to True, 0
            corresponds to False.

            resolution (int): Resolution is the starting resolution of the
            data being uploaded to the channel.

            windowrange (int, int): Window range is the maximum and minimum
            pixel values for a particular image. This is used so that the
            image can be displayed in a readable way for viewing through
            RESTful calls

            readonly (int): This option allows the user to control if,
            after the initial data commit, the channel is read-only.
            Generally this is suggested with data that will be publicly
            viewable.

        Returns:
            None
        """
        self.channels[
            channel_name] = [channel_name, datatype, channel_type, data_url,
                             file_format, file_type, exceptions, resolution,
                             windowrange, readonly]

    def add_project(self, project_name, token_name='', public=0):
        """
        Arguments:
            project_name (str): Project name is the specific project within
            a dataset’s name. If there is only one project associated with
            a dataset then standard convention is to name the project the
            same as its associated dataset.

            token_name (str): The token name is the default token. If you
            do not wish to specify one, a default one will be created for
            you with the same name as the project name. However, if the
            project is private you must specify a token.

            public (int): This option allows users to specify if they want
            the project/channels to be publicly viewable/search-able.
            (1, 0) = (TRUE, FALSE)

        Returns:
            None
        """

        self.project = (project_name, token_name, public)

    def add_dataset(self, dataset_name, imagesize, voxelres,
                    offset=(0, 0, 0), timerange=(0, 0), scalinglevels=0,
                    scaling=0):
        self.dataset = (dataset_name, imagesize, voxelres,
                        offset, timerange, scalinglevels, scaling)
        """
        Arguments:
            dataset_name (str): Dataset Name is the overarching name of the
            research effort. Standard naming convention is to do
            LabNamePublicationYear or LeadResearcherCurrentYear.

            imagesize (int, int, int): Image size is the pixel count
            dimensions of the data. For example is the data is stored as a
            series of 100 slices each 2100x2000 pixel TIFF images, the X,Y,Z
            dimensions are (2100, 2000, 100).

            voxelres (flt, flt, flt): Voxel Resolution is the number of
            voxels per unit pixel. We store X,Y,Z voxel resolution separately.

            offset (int, int, int): If your data is not well aligned and
            there is “excess” image data you do not wish to examine, but
            are present in your images, offset is how you specify where your
            actual image starts. Offset is provided a pixel coordinate offset
            from origin which specifies the “actual” origin of the image.
            The offset is for X,Y,Z dimensions.

            timerange (int, int): Time Range is a parameter to support
            storage of Time Series data, so the value of the tuple is a 0
            to X range of how many images over time were taken. It takes 2
            inputs timeStepStart and timeStepStop.

            scalinglevels (int): Scaling levels is the number of levels the
            data is scalable to (how many zoom levels are present in the
            data). The highest resolution of the data is at scaling level 0,
            and for each level up the data is down sampled by 2x2
            (per slice). To learn more about the sampling service used,
            visit the the propagation service page.

            scaling (int): Scaling is the orientation of the data being
            stored, 0 corresponds to a Z-slice orientation (as in a
            collection of tiff images in which each tiff is a slice on the
            z plane) and 1 corresponds to an isotropic orientation (in
            which each tiff is a slice on the y plane).

        Returns:
            None
        """

    def add_metadata(self, metadata=""):
        self.metadata = metadata

        """
        Arguements:
            metadata(str): Any metadata as appropriate from the LIMS schema

        Returns:
            None
        """

    def ocp_json(self, dataset, project, channel_list, metadata):
        """Genarate OCP json object"""
        ocp_dict = {}
        ocp_dict['dataset'] = self.dataset_dict(*dataset)
        ocp_dict['project'] = self.project_dict(*project)
        ocp_dict['metadata'] = metadata
        ocp_dict['channels'] = {}
        for channel_name, value in channel_list.iteritems():
            ocp_dict['channels'][channel_name] = self.channel_dict(*value)

        return json.dumps(ocp_dict, sort_keys=True, indent=4)

    def dataset_dict(
        self, dataset_name, imagesize, voxelres,
            offset=[0, 0, 0], timerange=[0, 0], scalinglevels=0, scaling=0):
        """Generate the dataset dictionary"""
        dataset_dict = {}
        dataset_dict['dataset_name'] = dataset_name
        dataset_dict['imagesize'] = imagesize
        dataset_dict['voxelres'] = voxelres
        if offset is not None:
            dataset_dict['offset'] = offset
        if timerange is not None:
            dataset_dict['timerange'] = timerange
        if scalinglevels is not None:
            dataset_dict['scalinglevels'] = scalinglevels
        if scaling is not None:
            dataset_dict['scaling'] = scaling
        return dataset_dict

    def channel_dict(
        self, channel_name, datatype, channel_type, data_url, file_format,
            file_type, exceptions=0, resolution=0,
            windowrange=[0, 0], readonly=0):
        """Genearte the project dictionary"""
        channel_dict = {}
        channel_dict['channel_name'] = channel_name
        channel_dict['datatype'] = datatype
        channel_dict['channel_type'] = channel_type
        if exceptions is not None:
            channel_dict['exceptions'] = exceptions
        if resolution is not None:
            channel_dict['resolution'] = resolution
        if windowrange is not None:
            channel_dict['windowrange'] = windowrange
        if readonly is not None:
            channel_dict['readonly'] = readonly
        channel_dict['data_url'] = data_url
        channel_dict['file_format'] = file_format
        channel_dict['file_type'] = file_type
        return channel_dict

    def project_dict(self, project_name, token_name='', public=0):
        """Genarate the project dictionary"""
        project_dict = {}
        project_dict['project_name'] = project_name
        if token_name is not None:
            project_dict['token_name'] = project_name \
                if token_name == '' else token_name
        if public is not None:
            project_dict['public'] = public
        return project_dict

    def verify_path(self, data):
        # Insert try and catch blocks
        try:
            token_name = data["project"]["token_name"]
        except:
            token_name = data["project"]["project_name"]

        channel_names = data["channels"].keys()

        for i in range(0, len(channel_names)):
            channel_type = data["channels"][
                channel_names[i]]["channel_type"]
            path = data["channels"][channel_names[i]]["data_url"]

            if (channel_type == "timeseries"):
                timerange = data["dataset"]["timerange"]
                for j in xrange(timerange[0], timerange[1] + 1):
                    # Test for tifs or such? Currently test for just not
                    # empty
                    work_path = "{}{}/{}/time{}/".format(
                        path, token_name, channel_names[i], j)
                    resp = requests.head(work_path)
                    assert(resp.status_code == 200)
            else:
                # Test for tifs or such? Currently test for just not empty
                work_path = "{}{}/{}/".format(
                    path, token_name, channel_names[i])
                resp = requests.head(work_path)
                print(work_path)
                assert(resp.status_code == 200)

    def put_data(self, data, site_host):
        # try to post data to the server
        URLPath = "{}ca/autoIngest/".format(site_host)
        try:
            r = requests.post(URLPath, data=data)
        except:
            print "Error in posting JSON file"

    def post_data(self, site_host):
        """
        Arguements:
            metadata(str): Any metadata as appropriate from the LIMS schema

        Returns:
            None
        """
        complete_example = (
            self.dataset, self.project, self.channels, self.metadata)
        data = self.ocp_json(*complete_example)

        self.verify_path(json.loads(data))
        self.put_data(data, siteHost)

    def output_json(self, file_name):
        complete_example = (
            self.dataset, self.project, self.channels, self.metadata)
        data = self.ocp_json(*complete_example)

        f = open(file_name, 'w')
        f.write(data)
        f.close()



import ndio

VERSION = ndio.version
"""
roll with
git tag VERSION
git push --tags
python setup.py sdist upload -r pypi
"""

from distutils.core import setup
setup(
    name = 'ndio',
    packages = [
        'ndio',
        'ndio.convert',
        'ndio.ramon',
        'ndio.remote',
        'ndio.utils'
    ],
    version = VERSION,
    description = 'A Python library for open neuroscience data access and manipulation.',
    author = 'Jordan Matelsky',
    author_email = 'jordan@neurodata.io',
    url = 'https://github.com/openconnectome/ndio',
    download_url = 'https://github.com/openconnectome/ndio/tarball/' + VERSION,
    keywords = [
        'brain',
        'medicine',
        'microscopy',
        'neuro',
        'neurodata',
        'neurodata.io',
        'neuroscience',
        'ndio',
        'ocpy',
        'ocp',
        'ocp.me',
        'connectome',
        'connectomics',
        'spatial',
        'EM',
        'MRI',
        'fMRI',
        'calcium'
    ],
    classifiers = [],
)
